function DemoPisarenkoHarmonicDecomposition()
% DemoPisarenkoHarmonicDecomposition
%
% Overview:
%   We use Pisarenko harmonic decomposition to determine the frequencies
%   in a signal.
%
% References:
%   Mathematical and Computational Methods in Photonics - Tutorial Notes
%
% Authors:
%   Habib Ammari, Brian Fitzpatrick, Matias Ruiz, Sanghyeon Yu.
clc
clear all
close all

tN = 200;
tMin = 0;
tMax = 10;
t = linspace(tMin, tMax, tN);
sFunc = @(f, t) exp(1i*2*pi*f*t);
sigmaFunc = @(P, f, m) P * sFunc(f, m);

P = [ 1, 1 ];
pN = length(P);

% --------------------------------------------------
% Create two complex exponentials that will be summed to
% give the signal.
% --------------------------------------------------
f = [ 0.3, 0.7 ];
showSignasl(sFunc, f, t);

s = zeros(pN, tN);
for ip=1:pN
    s(ip, :) = sFunc(f(ip), t);
end

% --------------------------------------------------
% Construct the autocovariance matrix for the signal.
% --------------------------------------------------
mN = 3;
sigma = zeros(mN, 1);
for m=0:mN-1
    for k=1:pN
        sigma(m+1) = sigma(m+1) + sigmaFunc(P(k), f(k), m);
    end
end

C = [ sigma(1) conj(sigma(2)) conj(sigma(3));...
    sigma(2) sigma(1) conj(sigma(2));...
    sigma(3) sigma(2) sigma(1); ];

% --------------------------------------------------
% Alternate representation of the covariance matrix
% --------------------------------------------------
% sVec = zeros(mN, pN);
% for k=1:pN
%     sVec(1:end, k) = [ sFunc(f(k), 0); sFunc(f(k), 1); sFunc(f(k), 2) ];
% end
%
% CAlt = zeros(size(C));
% for k=1:pN
%     CAlt = CAlt + P(k) * sVec(:,k) * sVec(:,k)';
% end

% --------------------------------------------------
% Generate an autocovariance matrix that includes noise.
% --------------------------------------------------
noise = 0.1;
CTotal = C + eye(size(C))*noise;

% --------------------------------------------------
% Take the eigendecomposition of the autocovariance matrix to obtain the
% signal subspace and the noise subspace.
% --------------------------------------------------
[V D] = eigs(CTotal);
[V, D]  = eig(CTotal);
D = diag(D);
[D,sidx] = sort(D, 'descend');
D = diag(D);
V = V(:,sidx);

CSignal = zeros(size(C));
for k=1:pN
    CSignal = CSignal + (eigs(k) + noise) * V(:,k) * V(:,k)';
end

CNoise = zeros(size(C));
for k=pN+1:mN
    CNoise = CNoise + noise * V(:,k) * V(:,k)';
end

% --------------------------------------------------
% Plot the signal subspace and the noise subspace.
% --------------------------------------------------
PlotSubspaces(V)

% --------------------------------------------------
% Generate test vectors over a range of frequencies and
% plot their projection against the noise subspace.
% --------------------------------------------------
freqMin = 1e-8;
freqMax = 2;
freqN = 500;
freqRange = linspace(freqMin, freqMax, freqN);

values = zeros((freqN), 1);
maxValue = 0;
for iFreq=1:freqN
    freq = freqRange(iFreq);
    sk = [ sFunc(freq, 0); sFunc(freq, 1); sFunc(freq, 2) ];
    values(iFreq) = abs(1/dot(sk,V(:, mN)));
    
    if values(iFreq) > maxValue
        maxValue = values(iFreq);
    end
    
    tol = 5e-2;
    if 1/values(iFreq) < tol
        col = 'r';
    else
        col = 'g';
    end
    
    skReal = real(sk);
    skImag = imag(sk);
    subplot(2, 2, 1);
    plot3(skReal(1), skReal(2), skReal(3),'.', 'MarkerEdgeColor', col,...
        'MarkerFaceColor', [.49 1 .63], 'MarkerSize', 10); title('Real')
    subplot(2, 2, 2);
    plot3(skImag(1), skImag(2), skImag(3),'.', 'MarkerEdgeColor', col,...
        'MarkerFaceColor', [.49 1 .63], 'MarkerSize', 10); title('Imaginary')
    
    subplot(2, 2, 3);
    plot(freqRange(1:iFreq), values(1:iFreq));
    axis([freqMin, freqMax, 0, 100])
    xlabel('Frequency')
    pause(0.1)
end

end

function showSignasl(sFunc, f, t)
h = figure(1);
set(h, 'Position', [200 100 1600 900])
subplot(2, 2, 1); s1 = sFunc(f(1), t); plot(t, real(s1)); title('s1'); xlabel('Time'); ylabel('Amplitude');
subplot(2, 2, 2); s2 = sFunc(f(2), t); plot(t, real(s2)); title('s2'); xlabel('Time'); ylabel('Amplitude');
subplot(2, 2, 3); sAll = s1 + s2; plot(t, real(sAll)); title('sAll'); xlabel('Time'); ylabel('Amplitude');
end

function PlotSubspaces(V)
drawArrow = @(x,y, color) quiver3(x(1),x(2),x(3),y(1),y(2),y(3), color) ;

aScale = 3;
h = figure(2);
set(h, 'Position', [200 100 1600 900])
subplot(2, 2, 1);
hold on
grid on
vScale = 2;
v1 = real(V(:,1) + V(:,2))*vScale;
v2 = real(V(:,1) - V(:,2))*vScale;
v3 = real(-V(:,1) - V(:,2))*vScale;
v4 = real(-V(:,1) + V(:,2))*vScale;

points=[v1 v2 v3 v4];
fill3(points(1,:), points(2,:), points(3,:),'k');
grid on
% alpha(0.2)
drawArrow([0,0,0], real(V(:,1)), 'k');
drawArrow([0,0,0], real(V(:,2)), 'k');
drawArrow([0,0,0], real(V(:,3)), 'r');
axis([-aScale, aScale, -aScale, aScale, -aScale, aScale])
rotate3d on
view([45, -80, 15]);

subplot(2, 2, 2);
hold on
grid on
vScale = 2;
v1 = real(V(:,1) + V(:,2))*vScale;
v2 = real(V(:,1) - V(:,2))*vScale;
v3 = real(-V(:,1) - V(:,2))*vScale;
v4 = real(-V(:,1) + V(:,2))*vScale;

points=[v1 v2 v3 v4];
fill3(points(1,:), points(2,:), points(3,:),'k');
grid on
% alpha(0.2)
drawArrow([0,0,0], real(V(:,1)), 'k');
drawArrow([0,0,0], real(V(:,2)), 'k');
drawArrow([0,0,0], real(V(:,3)), 'r');
axis([-aScale, aScale, -aScale, aScale, -aScale, aScale])
rotate3d on
view([45, -60, 15]);
end